#!/usr/bin/env python

# set env vars (optional):
# export PLAYSTORE_TOKEN='ya29.fooooo'
# export PLAYSTORE_GSFID='1234567891234567890'
# export HTTP_PROXY='http://localhost:8080'
# export HTTPS_PROXY='http://localhost:8080'
# export CURL_CA_BUNDLE='/usr/local/myproxy_info/cacert.pem'

from gplaycrawler.main import main


if __name__ == '__main__':
    main()



# stdlib
from datetime import datetime
import logging

# external
import coloredlogs  # colored logs


def get_logger(log_level, name=__name__):
    """
    Colored logging

    :param log_level:  'warning', 'info', 'debug'
    :param name: logger name (use __name__ variable)
    :return: Logger
    """

    fmt = '%(asctime)s %(threadName)-10s %(levelname)-8s %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'

    fs = {
        'asctime': {'color': 'green'},
        'hostname': {'color': 'magenta'},
        'levelname': {'color': 'red', 'bold': True},
        'name': {'color': 'magenta'},
        'programname': {'color': 'cyan'},
        'username': {'color': 'yellow'},
    }

    ls = {
        'critical': {'color': 'red', 'bold': True},
        'debug': {'color': 'green'},
        'error': {'color': 'red'},
        'info': {},
        'notice': {'color': 'magenta'},
        'spam': {'color': 'green', 'faint': True},
        'success': {'color': 'green', 'bold': True},
        'verbose': {'color': 'blue'},
        'warning': {'color': 'yellow'},
    }

    logger = logging.getLogger(name)

    # log to file
    now_str = datetime.now().strftime('%F_%T')
    handler = logging.FileHandler(f'{now_str}.log')
    formatter = logging.Formatter(fmt, datefmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # logger.propagate = False  # no logging of libs
    coloredlogs.install(level=log_level, logger=logger, fmt=fmt, datefmt=datefmt, level_styles=ls, field_styles=fs)
    return logger



from os import path, mkdir, walk
from requests.exceptions import HTTPError, ReadTimeout, ChunkedEncodingError
import json
import queue
import threading
import time

from playstoreapi.googleplay import GooglePlayAPI, LoginError, RequestError
from gplaycrawler.utils import get_logger


class Packages:
    def __init__(self, locale='en_US', timezone='UTC', device='px_3a', delay=None, log_level='info'):
        self.locale = locale
        self.timezone = timezone
        self.device = device
        self.delay = delay
        self.lastRequest = 0

        self.log_level = log_level
        self.log = get_logger(log_level, name=__name__)
        self.log.debug('Logging is set to debug')

        if log_level.lower() == 'debug':
            self.quiet = False
        else:
            self.quiet = True

    def worker(self, api, q, done_ids, out_dir, threads, expansion_files, splits):
        while not q.empty():
            if self.delay:
                tosleep = -(time.time() - self.lastRequest - self.delay)
                if tosleep > 0:
                    # print('sleeping', tosleep)
                    time.sleep(tosleep)

            packageName = q.get_nowait()  # non blocking
            self.log.debug(f'Start getting package "{packageName}"')

            try:
                # versionCode=None, offerType=1, expansion_files=False
                download = api.download(packageName, expansion_files=expansion_files)
                self.lastRequest = time.time()
            except RequestError as e:
                if "Can't install. Please try again later." in str(e):
                    self.log.info(f'{packageName}: paid app. Skipping')
                else:
                    self.log.warning(f'{packageName}: {str(e)}')
                    self.log.debug('Readd to queue')
                    q.put_nowait(packageName)
                q.task_done()
                continue
            except HTTPError as e:
                if e.response.status_code == 429 or e.response.status_code == 401:
                    if e.response.status_code == 401:
                        self.log.warning('Unauthorized. Trying relogin...')
                    else:
                        self.log.warning('packages got rate limited')
                    api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                    while True:
                        try:
                            api.envLogin(quiet=self.quiet, check=False)
                        except LoginError:
                            self.log.info('Login failed. Retry in 3 minutes')
                            time.sleep(180)
                        else:
                            break
                    self.log.debug('new api, logged in, try again')
                else:
                    self.log.warning(str(e))
                    q.task_done()
                    continue
            except ReadTimeout:
                self.log.debug(f'Request did timeout (worker), add back {packageName}')
                q.put_nowait(packageName)
                q.task_done()
                continue

            # download = server.download(docid, expansion_files=True)

            if download['docId'] != packageName:
                self.log.warning(f"package name doesn't match {download['docId']} != {packageName}")

            filepath = path.join(out_dir, packageName + '.apk')

            self.log.debug('Downloading apk...')
            try:
                with open(filepath, 'wb') as f:
                    for chunk in download.get('file').get('data'):
                        f.write(chunk)
            except ChunkedEncodingError as e:
                self.warning(f'Error writing chunk for {packageName}.apk: {str(e)}')

            if expansion_files is True and download['additionalData'] != []:
                self.log.debug('Downloading additional files...')
                for obb in download['additionalData']:
                    name = f"{packageName}.{obb['type']}.{str(obb['versionCode'])}.obb"
                    filepath = path.join(out_dir, name)
                    self.log.debug(f'Additional file: {name}')
                    try:
                        with open(filepath, 'wb') as f:
                            for chunk in obb.get('file').get('data'):
                                f.write(chunk)
                    except ChunkedEncodingError as e:
                        self.warning(f'Error writing chunk for {name}: {str(e)}')

            if splits is True and download['splits'] != []:
                self.log.debug('Downloading splits...')
                for split in download.get('splits'):
                    name = f"{packageName}.split.{split['name']}.zip"
                    filepath = path.join(out_dir, name)
                    self.log.debug(f'Split: {name}')
                    try:
                        with open(filepath, "wb") as f:
                            for chunk in split["file"]["data"]:
                                f.write(chunk)
                    except ChunkedEncodingError as e:
                        self.warning(f'Error writing chunk for {name}: {str(e)}')

            q.task_done()
            self.log.info(f'Done: {len(done_ids)}, to do: {q.qsize()}')
            done_ids.add(packageName)

        t = threading.current_thread()
        self.log.info(f'{t.name} finished. Queue empty')
        t.done = True
        return

    def getPackages(self, in_file, out_dir, num_threads, expansion_files, splits):
        ids = set()

        try:
            with open(in_file) as f:
                j = json.load(f)
        except FileNotFoundError:
            self.log.error('Input file not found')
            return

        if type(j) == list:
            ids.update(j)
        else:
            try:
                ids.update(j['done'])
                ids.update(j['ids'])
            except KeyError:
                self.log.error('Input file is in unknown format')

        # load ids that are already done via amp
        try:
            (_, _, done_filenames) = next(walk(out_dir))
        except StopIteration:
            done_filenames = []

        ids_done = set()
        for file in set(done_filenames):
            ids_done.add(file.rstrip('.apk'))

        todo_ids = ids - ids_done
        q = queue.Queue()
        for packageName in todo_ids:
            q.put_nowait(packageName)

        self.log.info(f'Done: {len(ids_done)}, to do: {q.qsize()}')
        try:
            mkdir(out_dir)
        except FileExistsError:
            pass

        i = 0
        threads = []
        while num_threads > 0:
            while len(threads) < num_threads:
                api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                while True:
                    try:
                        api.envLogin(quiet=self.quiet, check=False)
                    except LoginError:
                        self.log.info('Login failed. Retry in 3 minutes')
                        time.sleep(180)
                    else:
                        break
                t = threading.Thread(
                    target=self.worker, args=(api, q, ids_done, out_dir, threads, expansion_files, splits)
                )
                i += 1
                t.name = f'Worker-{i}'
                t.done = False
                threads.append(t)
                t.start()
                self.log.info(f'Started {t.name}')
                # wait 10-11 min to start next thread
                # time.sleep(60 * 10 + random() * 60)

            for t in threads:
                t.join(timeout=1)
                if not t.is_alive():
                    if t.done:
                        num_threads -= 1
                    else:
                        threads.remove(t)
                        self.log.info(f'Worker {t.name} crashed. Starting a new worker')
                        time.sleep(1)

        self.log.info('Threads finished')



import json
import threading
import time
import queue
from requests.exceptions import HTTPError, ReadTimeout

from playstoreapi.googleplay import GooglePlayAPI, LoginError
from gplaycrawler.utils import get_logger

import string
import itertools


class Search:
    def __init__(self, locale='en_US', timezone='UTC', device='px_3a', delay=None, log_level='info'):
        self.locale = locale
        self.timezone = timezone
        self.device = device
        self.delay = delay

        self.log_level = log_level
        self.log = get_logger(log_level, name=__name__)
        self.log.debug('Logging is set to debug')

        if log_level.lower() == 'debug':
            self.quiet = False
        else:
            self.quiet = True

    def search(self, api, nextPageUrl=None, query=None):
        np = queue.Queue()
        if query is None:
            np.put_nowait(nextPageUrl)

        ids = set()
        done = set()
        num_pages = 0
        while not np.empty() or query is not None:
            nextPageUrl = None
            if query is None:
                nextPageUrl = np.get_nowait()
                if nextPageUrl in done:
                    continue
                done.add(nextPageUrl)

            try:
                result = api.search(query=query, nextPageUrl=nextPageUrl)
                query = None
            except HTTPError as e:
                if e.response.status_code == 429 or e.response.status_code == 401:
                    if e.response.status_code == 401:
                        self.log.warning('Unauthorized. Trying relogin...')
                    else:
                        self.log.warning('search got rate limited')
                    api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                    while True:
                        try:
                            api.envLogin(quiet=self.quiet, check=False)
                        except LoginError:
                            self.log.info('Login failed. Retry in 3 minutes')
                            time.sleep(180)
                        else:
                            break
                    self.log.debug('new api, logged in, try again')
                    np.put_nowait(nextPageUrl)
                else:
                    self.log.warning(str(e))
                continue

            if len(result) != 1:
                self.log.warning(f'Pages: Got result with len {len(result)}')
                self.log.debug(result)
                continue

            doc = result[0]
            clusters = doc.get('subItem')

            try:
                np.put_nowait(doc['containerMetadata']['nextPageUrl'])
            except KeyError:
                pass

            if clusters is None:
                self.log.debug('No clusters found in searchPages')
                continue

            num_pages += 1

            # at the first page there are also the clusters 'Recommended for you' and 'Related Searches'
            # so the item can be list of apps or related search terms
            for cluster in clusters:
                apps = 0
                try:
                    items = cluster['subItem']
                except KeyError:
                    items = []
                for item in items:
                    try:
                        ids.add(item['id'])
                    except KeyError:
                        pass
                    else:
                        apps += 1
                self.log.debug(f"queue: {np.qsize()}, pages: Found {apps} apps in cluster {cluster.get('title')}")

                try:
                    np.put_nowait(cluster['containerMetadata']['nextPageUrl'])
                except KeyError:
                    pass
        return ids

    def worker(self, api, q, ids, done_searchTerms, tmp_file, threads):
        while not q.empty():
            searchTerm = q.get_nowait()  # non blocking
            self.log.debug(f'Start searching "{searchTerm}"')

            try:
                sids = self.search(api, query=searchTerm)

            except HTTPError as e:
                if e.response.status_code == 429 or e.response.status_code == 401:
                    if e.response.status_code == 401:
                        self.log.warning('Unauthorized. Trying relogin...')
                    else:
                        self.log.warning('search worker got rate limited')
                    api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                    while True:
                        try:
                            api.envLogin(quiet=self.quiet, check=False)
                        except LoginError:
                            self.log.info('Login failed. Retry in 3 minutes')
                            time.sleep(180)
                        else:
                            break
                    self.log.debug(f'new api, logged in, add back {searchTerm}')
                    q.put_nowait(searchTerm)
                    q.task_done()
                    continue
                else:
                    self.log.warning(str(e))

            except ReadTimeout:
                self.log.debug(f'Request did timeout (search), add back {searchTerm}')
                q.put_nowait(searchTerm)
                q.task_done()
                continue

            q.task_done()
            before = len(ids)
            ids.update(sids)
            new = len(ids) - before
            self.log.debug(f'{searchTerm}: Got {len(sids)} package names, {new} new')

            self.log.info(f'Done: {len(done_searchTerms)}, to do: {q.qsize()} received {len(ids)} ids')
            done_searchTerms.add(searchTerm)

            # save tmp file regularly
            if len(done_searchTerms) % (100 * len(threads)) == 0:
                self.log.debug(f'Saving tmp file {tmp_file}')
                with open(tmp_file, 'w') as f:
                    json.dump({"done": list(done_searchTerms), "ids": list(ids)}, f, indent=2)

        t = threading.current_thread()
        self.log.info(f'{t.name} finished. Queue empty')
        t.done = True
        return

    def get_strings(self, alphabet):
        with open('chars.json') as f:
            chars = json.load(f)
        for charset in chars:
            if charset['name'] == alphabet:
                return charset['chars']

    def generate_strings(self, length=3):
        chars = string.ascii_lowercase
        # nums = string.digits

        for item in itertools.product(chars, repeat=length):
            yield "".join(item)

    def getSearch(self, out_file, num_threads, length):
        all_searchTerms = set()
        for s in self.generate_strings(length=length):
            all_searchTerms.add(s.lower())

        out_file = out_file.rstrip('.json')

        tmp_file = out_file + '_tmp.json'

        ids = set()
        done_searchTerms = set()

        # resume from tmp file if it exists
        try:
            with open(tmp_file) as f:
                done = json.load(f)
            done_searchTerms.update(done['done'])
            ids.update(done['ids'])
        except FileNotFoundError:
            pass
        except Exception as e:
            self.log.warning(str(e))

        todo_seachTerms = all_searchTerms - done_searchTerms
        q = queue.Queue()
        for searchTerm in todo_seachTerms:
            q.put_nowait(searchTerm)

        self.log.info(f'Done: {len(done_searchTerms)}, to do: {q.qsize()} received {len(ids)} ids')

        i = 0
        threads = []
        while num_threads > 0:
            while len(threads) < num_threads:
                api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                while True:
                    try:
                        api.envLogin(quiet=self.quiet, check=False)
                    except LoginError:
                        self.log.info('Login failed. Retry in 3 minutes')
                        time.sleep(180)
                    else:
                        break
                t = threading.Thread(target=self.worker, args=(api, q, ids, done_searchTerms, tmp_file, threads))
                i += 1
                t.name = f'Worker-{i}'
                t.done = False
                threads.append(t)
                t.start()
                self.log.info(f'Started {t.name}')
                # wait 10-11 min to start next thread
                # time.sleep(60 * 10 + random() * 60)

            for t in threads:
                t.join(timeout=1)
                if not t.is_alive():
                    if t.done:
                        num_threads -= 1
                    else:
                        threads.remove(t)
                        self.log.info(f'Worker {t.name} crashed. Starting a new worker')
                        time.sleep(1)

        self.log.info('Threads finished')
        self.log.info(f'Saving out file {out_file}.json')
        with open(out_file + '.json', 'w') as f:
            json.dump({"done": list(done_searchTerms), "ids": list(ids)}, f, indent=2)



import json
import threading
import time
import queue
from requests.exceptions import HTTPError, ReadTimeout
import random

from playstoreapi.googleplay import GooglePlayAPI, LoginError
from gplaycrawler.utils import get_logger


# def details(packageName):

#     d = api.details(packageName)
#     filename = packageName + '.json'
#     with open(filename, 'w') as f:
#         json.dump(d, f)

#     d = api.bulkDetails([packageName])
#     filename = packageName + '_bulk.json'
#     with open(filename, 'w') as f:
#         json.dump(d, f)


class Related:
    def __init__(self, locale='en_US', timezone='UTC', device='px_3a', delay=None, log_level='info'):
        self.locale = locale
        self.timezone = timezone
        self.device = device
        self.delay = delay

        self.log_level = log_level
        self.log = get_logger(log_level, name=__name__)
        self.log.debug('Logging is set to debug')

        if log_level.lower() == 'debug':
            self.quiet = False
        else:
            self.quiet = True

    def streamPages(self, api, nextPageUrl):
        ids = set()
        while True:
            try:
                d = api.streamDetails(nextPageUrl=nextPageUrl)
            except HTTPError as e:
                if e.response.status_code == 429 or e.response.status_code == 401:
                    if e.response.status_code == 401:
                        self.log.warning('Unauthorized. Trying relogin...')
                    else:
                        self.log.debug('streamPages got rate limited')
                    api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                    while True:
                        try:
                            api.envLogin(quiet=self.quiet)
                        except LoginError:
                            self.log.info('Login failed. Retry in 3 minutes')
                            time.sleep(180)
                        else:
                            break
                    self.log.debug('new api, logged in, try again')
                else:
                    self.log.warning(str(e))
                continue

            except ReadTimeout:
                self.log.debug('Request did timeout (pages), try again in 1 min')
                time.sleep(60)
                continue
            try:
                streamBundle = d['item'][0]['subItem']
            except KeyError:
                try:
                    streamBundle = d[0]['subItem']
                except KeyError:
                    self.log.debug(f'No bundle found (pages) {d}')
                    return ids
            if len(streamBundle) > 1:
                self.log.warning('This should not happen')
                return ids
            stream = streamBundle[0]
            try:
                nextPageUrl = stream['containerMetadata']['nextPageUrl']
            except KeyError:
                return ids

            for subItem in stream['subItem']:
                ids.add(subItem['id'])
            # print(stream['title'], len(ids))

    def streamDetails(self, api, ids, packageName):
        d = api.streamDetails(packageName)
        try:
            streamBundle = d['item'][0]['subItem']
        except KeyError:
            try:
                streamBundle = d[0]['subItem']
            except KeyError:
                self.log.warning(f'{packageName}: No bundle found (details)')
                self.log.debug(d)
                return
        for stream in streamBundle:
            # print(packageName, stream['title'])
            for subItem in stream['subItem']:
                ids.add(subItem['id'])

            # get all pages for every stream
            try:
                nextPageUrl = stream['containerMetadata']['nextPageUrl']
            except KeyError:
                pass
            else:
                i = self.streamPages(api, nextPageUrl)
                before = len(ids)
                ids.update(i)
                new = len(ids) - before
                self.log.debug(f"{packageName} {stream['title']}: Got {len(i)} package names, {new} new")

            # print('return', len(ids), ids)
            # return ids

    def worker(self, api, q, ids, done_ids, shared, tmp_file, out_file, until_level=3):
        s = shared
        t = threading.current_thread()
        t.level = s['level']
        t.idle = False
        while True:
            try:
                packageName = q.get_nowait()  # non blocking
                self.log.debug(f'Start crawling {packageName}')
            except queue.Empty:
                t.idle = True
                time.sleep(random.random())
                if t.level != s['level']:
                    if s['level'] > until_level:
                        self.log.info('Done')
                        t.done = True
                        return
                    else:
                        self.log.debug('sleeping (wait for save and filled queue')
                        while q.empty():
                            time.sleep(1)
                        t.idle = False
                        t.level += 1
                        continue

                # this code is just run by the first thread
                # level up and save
                t.level += 1
                s['level'] += 1
                for thread in s['threads']:
                    while not thread.idle:
                        self.log.debug('Waiting for other threads to finish level')
                        time.sleep(1)

                todo_ids = ids - done_ids
                self.log.info(f'Done with level {s["level"]}. Got {len(ids)} IDs. Done IDs: {len(done_ids)}')

                self.log.info('Saving json file\n')
                if s['level'] > until_level:
                    filename = f'{out_file}.json'
                else:
                    filename = f'{out_file}_level-{s["level"]}.json'

                with open(filename, 'w') as f:
                    json.dump({"done": list(done_ids), "ids": list(ids)}, f, indent=2)

                if s['level'] > until_level:
                    self.log.info(f'{t.name}: First thread done')
                    t.done = True
                    return

                for packageName in todo_ids:
                    q.put_nowait(packageName)

                self.log.info(f'Filled queue for next level: {s["level"]}, Todo: {q.qsize()}\n')
                continue
            try:
                self.streamDetails(api, ids, packageName)
            except HTTPError as e:
                if e.response.status_code == 429 or e.response.status_code == 401:
                    if e.response.status_code == 401:
                        self.log.warning('Unauthorized. Trying relogin...')
                    else:
                        self.log.debug('streamDetails got rate limited')
                    api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                    while True:
                        try:
                            api.envLogin(quiet=self.quiet)
                        except LoginError:
                            self.log.info('Login failed. Retry in 3 minutes')
                            time.sleep(180)
                        else:
                            break
                    self.log.debug(f'new api, logged in, add back {packageName}')
                    q.put_nowait(packageName)
                    q.task_done()
                    continue
                else:
                    self.log.warning(str(e))
            except ReadTimeout:
                self.log.debug(f'Request did timeout (details), add back {packageName}')
                q.put_nowait(packageName)
                q.task_done()
                continue

            q.task_done()
            self.log.info(f'Done: {len(done_ids)}, to do: {q.qsize()} received {len(ids)} ids (level {t.level})')
            done_ids.add(packageName)

            # save tmp file regularly
            num_threads = len(s['threads'])
            if len(done_ids) % (100 * num_threads) == 0:
                self.log.debug(f'Saving tmp file {tmp_file}')
                with open(tmp_file, 'w') as f:
                    json.dump({"done": list(done_ids), "ids": list(ids)}, f, indent=2)

    def getRelated(self, in_file, out_file, until_level=3, threads=3):
        '''
        parallel downloading of related apps from charts.json and saving them in a json file
        '''
        with open(in_file) as f:
            input_dict = json.load(f)

        in_ids = set()
        try:
            for cat in input_dict:
                for chart in input_dict[cat]:
                    in_ids.update(input_dict[cat][chart])
        except TypeError:
            self.log.info('doesn\'t look like a charts.json. Trying to load level file')
            if type(input_dict) == list:
                in_ids.update(input_dict)
            elif type(input_dict.get('ids')) == list:
                in_ids.update(input_dict['ids'])
            else:
                self.log.error('Could not load input file')
                return

        out_file = out_file.rstrip('.json')

        ids = set()
        done_ids = set()

        # resume from tmp file if it exists
        tmp_file = out_file + '_tmp.json'
        try:
            with open(tmp_file) as f:
                done = json.load(f)
            done_ids.update(done['done'])
            ids.update(done['ids'])
        except FileNotFoundError:
            pass
        except Exception as e:
            self.log.warning(str(e))

        todo_ids = in_ids - done_ids
        q = queue.Queue()
        for packageName in todo_ids:
            q.put_nowait(packageName)

        self.log.info(f'Done: {len(done_ids)}, to do: {q.qsize()} received {len(ids)} ids')

        # needs to be dict
        shared = dict()
        shared['level'] = 0
        shared['threads'] = []

        i = 0
        while True:
            while len(shared['threads']) < threads:
                api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
                while True:
                    try:
                        api.envLogin(quiet=self.quiet)
                    except LoginError:
                        self.log.info('Login failed. Retry in 3 minutes')
                        time.sleep(180)
                    else:
                        break

                t = threading.Thread(
                    target=self.worker, args=(api, q, ids, done_ids, shared, tmp_file, out_file, until_level)
                )
                i += 1
                t.name = f'Worker-{i}'
                t.done = False
                shared['threads'].append(t)
                t.start()
                self.log.info(f'Started {t.name}')

            for t in shared['threads']:
                t.join(timeout=1)
                if not t.is_alive():
                    if t.done:
                        threads -= 1
                    else:
                        self.log.info(f'Worker {t.name} crashed. Starting a new worker')
                        time.sleep(1)

        self.log.info('Threads finished')
        self.log.info(f'Saving out file {out_file}.json')
        with open(out_file + '.json', 'w') as f:
            json.dump({"done": list(done_ids), "ids": list(ids)}, f, indent=2)



import json
import threading
import time

from reprint import output
from playstoreapi.googleplay import GooglePlayAPI, LoginError
from gplaycrawler.utils import get_logger


class Charts:
    def __init__(self, locale='en_US', timezone='UTC', device='px_3a', delay=None, log_level='info'):
        self.locale = locale
        self.timezone = timezone
        self.device = device
        self.delay = delay

        self.log_level = log_level
        self.log = get_logger(log_level, name=__name__)
        self.log.debug('Logging is set to debug')

        if log_level.lower() == 'debug':
            self.quiet = False
        else:
            self.quiet = True

    def getChart(self, api, categorie, chart, ids):
        '''
        Get all ids by downloading all pages
        Every page has 6 entries and the api returns ~110 pages resulting in ~660 ids
        '''
        if categorie not in ids:
            ids[categorie] = {}
        if chart not in ids[categorie]:
            ids[categorie][chart] = []
        nextPageUrl = None
        while True:
            data = api.topChart(cat=categorie, chart=chart, nextPageUrl=nextPageUrl)
            s = data.get('subItem')
            if s is None:
                self.log.warning('No subItem found')
                self.log.debug(s)
                continue
            for subItem in data['subItem']:
                for app in subItem['subItem']:
                    # print('\t{} ({}): {}'.format(chart, len(ids[chart]), app['id']))
                    ids[categorie][chart].append(app['id'])
                try:
                    nextPageUrl = subItem['containerMetadata']['nextPageUrl']
                except KeyError:
                    return

    def getCharts(self, out_file):
        '''
        parallel downloading of all app charts and saving them in a json file
        '''
        api = GooglePlayAPI(self.locale, self.timezone, self.device, delay=self.delay)
        try:
            api.envLogin(quiet=self.quiet, check=False)
        except LoginError:
            self.log.info('Login failed.')

        charts = ['apps_topselling_free', 'apps_topselling_paid', 'apps_topgrossing', 'apps_movers_shakers']
        categories = ['APPLICATION', 'GAME']
        ids = {}
        threads = []
        for cat in categories:
            for chart in charts:
                t = threading.Thread(target=self.getChart, args=(api, cat, chart, ids))
                t.name = f'{cat}: {chart}'
                threads.append(t)
                print('starting thread', t.name)
                t.start()

        with output(output_type='dict') as output_lines:
            alive = 1
            while alive > 0:
                alive = 0
                for t in threads:
                    n = t.name.split(': ')
                    cat = n[0]
                    chart = n[1]
                    if t.is_alive():
                        alive += 1
                        output_lines[t.name] = f'{len(ids[cat][chart])} ...'
                    else:
                        output_lines[t.name] = f'{len(ids[cat][chart])} (done)'
                time.sleep(0.1)

        self.log.info('Waiting for threads')
        for t in threads:
            t.join()

        self.log.info(f'Writing to file {out_file}')
        with open(out_file, 'w') as fp:
            json.dump(ids, fp, indent=2)



# stdlib
from argparse import ArgumentParser, HelpFormatter
from importlib.metadata import metadata

# internal
from gplaycrawler.related import Related
from gplaycrawler.search import Search
from gplaycrawler.metadata import Metadata
from gplaycrawler.packages import Packages
from gplaycrawler.charts import Charts


class F(HelpFormatter):
    def __init__(self, *args, **kwargs):
        kwargs['max_help_position'] = 30
        super().__init__(*args, **kwargs)


def main():
    parser = ArgumentParser(description=metadata(__package__)['Summary'])
    parser.add_argument(
        '-v',
        '--verbosity',
        help='Set verbosity level (default: %(default)s)',
        choices=['warning', 'info', 'debug'],
        default='info',
    )
    subparsers = parser.add_subparsers(help='Desired action to perform', dest='command')

    # help
    subparsers.add_parser('help', help='Print this help message')

    # usage
    subparsers.add_parser('usage', help='Print full usage')

    # Create parent subparser for with common arguments
    parent_parser = ArgumentParser(add_help=False, formatter_class=F)
    parent_parser.add_argument('--locale', default='en_US', help='(default: %(default)s)')
    parent_parser.add_argument('--timezone', default='UTC', help='(default: %(default)s)')
    parent_parser.add_argument('--device', default='px_3a', help='(default: %(default)s)')
    parent_parser.add_argument(
        '--delay', default=0.51, help='Delay between every request in seconds (default: %(default)s)', type=float
    )
    parent_parser.add_argument(
        '--threads', default=2, help='Number of parallel workers (default: %(default)s)', type=int
    )

    # Subparsers based on parent

    # charts
    d = 'parallel downloading of all cross category app charts'
    parser_charts = subparsers.add_parser('charts', parents=[parent_parser], help=d, description=d, formatter_class=F)

    parser_charts.add_argument('--output', help='name of the output file (default: %(default)s)', default='charts.json')

    # search
    d = 'parallel searching of apps via search terms'
    parser_search = subparsers.add_parser('search', parents=[parent_parser], help=d, description=d, formatter_class=F)

    parser_search.add_argument(
        '--output', help='name of the output file (default: %(default)s)', default='ids_search.json'
    )
    parser_search.add_argument(
        '--length', default=2, help='length of strings to search (default: %(default)s)', type=int
    )

    # related
    d = 'parallel searching of apps via related apps'
    parser_related = subparsers.add_parser('related', parents=[parent_parser], help=d, description=d, formatter_class=F)

    parser_related.add_argument('input', help='name of the input file (default: %(default)s)', default='charts.json')
    parser_related.add_argument(
        '--output', help='base name of the output files (default: %(default)s)', default='ids_related'
    )
    parser_related.add_argument('--level', default=3, help='How deep to crawl (default: %(default)s)', type=int)

    # metadata
    d = 'parallel scraping of app metadata'
    parser_metadata = subparsers.add_parser(
        'metadata', parents=[parent_parser], help=d, description=d, formatter_class=F
    )

    parser_metadata.add_argument('input', help='name of the input file (json)')
    parser_metadata.add_argument(
        '--output', help='directory name of the output files (default: %(default)s)', default='out_metadata'
    )
    # packages
    d = 'parallel downloading app packages'
    parser_packages = subparsers.add_parser(
        'packages', parents=[parent_parser], help=d, description=d, formatter_class=F
    )

    parser_packages.add_argument('input', help='name of the input file (json)')
    parser_packages.add_argument(
        '--output', help='directory name of the output files (default: %(default)s)', default='out_packages'
    )
    parser_packages.add_argument(
        '--expansions', help='also download expansion files (default: %(default)s)', action='store_true'
    )
    parser_packages.add_argument(
        '--splits', help='also download split files (default: %(default)s)', action='store_true'
    )

    args = parser.parse_args()

    if args.command == 'help' or args.command is None:
        parser.print_help()
        exit()

    if args.command == 'usage':
        parser.print_help()
        print('\n\nAll commands in detail:')

        parentsubparsers = [parser_related, parser_search, parser_metadata, parser_packages]
        commonargs = ['-h, --help', '--locale', '--timezone', '--device', '--delay', '--threads']
        parentsubparsers_str = []
        for p in parentsubparsers:
            parentsubparsers_str.append(p.prog.split(' ')[1])

        print(f'\n\nCommon optional arguments for {", ".join(parentsubparsers_str)}:')
        print('\n'.join(parent_parser.format_help().splitlines()[4:]))
        for p, p_str in zip(parentsubparsers, parentsubparsers_str):
            h = p.format_help()
            hn = ''
            for line in h.splitlines():
                add = True
                for arg in commonargs:
                    if line.lstrip().startswith(arg):
                        add = False
                if add:
                    hn += line + '\n'
            hn = hn.rstrip('optional arguments:\n')
            print(f"\n\n{p_str}:\n{hn}")
        exit()

    if args.command == 'charts':
        c = Charts(
            locale=args.locale, timezone=args.timezone, device=args.device, delay=args.delay, log_level=args.verbosity
        )
        c.getCharts(args.output)
    elif args.command == 'search':
        s = Search(
            locale=args.locale, timezone=args.timezone, device=args.device, delay=args.delay, log_level=args.verbosity
        )
        s.getSearch(args.output, args.threads, args.length)
    elif args.command == 'related':
        r = Related(
            locale=args.locale, timezone=args.timezone, device=args.device, delay=args.delay, log_level=args.verbosity
        )
        r.getRelated(args.input, args.output, until_level=args.level, threads=args.threads)
    elif args.command == 'metadata':
        m = Metadata(
            locale=args.locale, timezone=args.timezone, device=args.device, delay=args.delay, log_level=args.verbosity
        )
        m.getMetadata(args.input, args.output, args.threads)
    elif args.command == 'packages':
        p = Packages(
            locale=args.locale, timezone=args.timezone, device=args.device, delay=args.delay, log_level=args.verbosity
        )
        p.getPackages(args.input, args.output, args.threads, args.expansions, args.splits)

    else:
        parser.print_help()
        exit()
